#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    routeSampler.py
# @author  Jakob Erdmann
# @date    2020-02-07

"""
Samples routes from a given set to fullfill specified counting data (edge counts or turn counts)
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import random
from collections import defaultdict
# multiprocessing imports
import multiprocessing
import numpy as np

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
import sumolib  # noqa
from sumolib.miscutils import parseTime  # noqa


def _run_func(args):
    func, interval, kwargs, num = args
    kwargs["cpuIndex"] = num
    return num, func(interval=interval, **kwargs)


def multi_process(cpu_num, seed, interval_list, func, outf, mismatchf, **kwargs):
    cpu_count = min(cpu_num, multiprocessing.cpu_count()-1)
    interval_split = np.array_split(interval_list, cpu_count)
    # pool = multiprocessing.Pool(processes=cpu_count)
    with multiprocessing.get_context("spawn").Pool() as pool:
        results = pool.map(_run_func, [(func, interval, kwargs, i) for i, interval in enumerate(interval_split)])
        # pool.close()
        results = sorted(results, key=lambda x: x[0])
        for _, result in results:
            outf.write("".join(result[-2]))
            if mismatchf is not None:
                mismatchf.write("".join(result[-1]))
        return results


def get_options(args=None):
    parser = sumolib.options.ArgumentParser(description="Sample routes to match counts")
    parser.add_argument("-r", "--route-files", dest="routeFiles",
                        help="Input route file")
    parser.add_argument("-t", "--turn-files", dest="turnFiles",
                        help="Input turn-count file")
    parser.add_argument("-d", "--edgedata-files", dest="edgeDataFiles",
                        help="Input edgeData file (for counts)")
    parser.add_argument("-O", "--od-files", dest="odFiles",
                        help="Input edgeRelation file for origin-destination counts")
    parser.add_argument("--edgedata-attribute", dest="edgeDataAttr", default="entered",
                        help="Read edgeData counts from the given attribute")
    parser.add_argument("--turn-attribute", dest="turnAttr", default="count",
                        help="Read turning counts from the given attribute")
    parser.add_argument("--turn-max-gap", type=int, dest="turnMaxGap", default=0,
                        help="Allow at most a gap of INT edges between from-edge and to-edge")
    parser.add_argument("-o", "--output-file", dest="out", default="out.rou.xml",
                        help="Output route file")
    parser.add_argument("--prefix", dest="prefix", default="",
                        help="prefix for the vehicle ids")
    parser.add_argument("-a", "--attributes", dest="vehattrs", default="",
                        help="additional vehicle attributes")
    parser.add_argument("-s", "--seed", type=int, default=42,
                        help="random seed")
    parser.add_argument("--mismatch-output", dest="mismatchOut",
                        help="write cout-data with overflow/underflow information to FILE")
    parser.add_argument("--weighted", dest="weighted", action="store_true", default=False,
                        help="Sample routes according to their probability (or count)")
    parser.add_argument("--optimize",
                        help="set optimization method level (full, INT boundary)")
    parser.add_argument("--optimize-input", dest="optimizeInput", action="store_true", default=False,
                        help="Skip resampling and run optimize directly on the input routes")
    parser.add_argument("--minimize-vehicles", dest="minimizeVehs", type=float, default=0,
                        help="Set optimization factor from [0, 1[ for reducing the number of vehicles"
                        + "(prefer routes that pass multiple counting locations over routes that pass fewer)")
    parser.add_argument("--geh-ok", dest="gehOk", default=5,
                        help="threshold for acceptable GEH values")
    parser.add_argument("-f", "--write-flows", dest="writeFlows",
                        help="write flows with the give style instead of vehicles [number|probability]")
    parser.add_argument("-I", "--write-route-ids", dest="writeRouteIDs", action="store_true", default=False,
                        help="write routes with ids")
    parser.add_argument("-u", "--write-route-distribution", dest="writeRouteDist",
                        help="write routeDistribution with the given ID instead of individual routes")
    parser.add_argument("-b", "--begin", help="custom begin time (seconds or H:M:S)")
    parser.add_argument("-e", "--end", help="custom end time (seconds or H:M:S)")
    parser.add_argument("-i", "--interval", help="custom aggregation interval (seconds or H:M:S)")
    parser.add_argument("-v", "--verbose", action="store_true", default=False,
                        help="tell me what you are doing")
    parser.add_argument("-V", "--verbose.histograms", dest="verboseHistogram", action="store_true", default=False,
                        help="print histograms of edge numbers and detector passing count")
    parser.add_argument("--threads", dest="threads", type=int, default=1,
                        help="If parallelization is desired, enter the number of CPUs to use. Set to a value >> then "
                             "your machines CPUs if you want to utilize all CPUs (Default is 1)"
                        )

    options = parser.parse_args(args=args)
    if (options.routeFiles is None or
            (options.turnFiles is None and options.edgeDataFiles is None and options.odFiles is None)):
        parser.print_help()
        sys.exit()
    if options.writeRouteIDs and options.writeRouteDist:
        sys.stderr.write("Only one of the options --write-route-ids and --write-route-distribution may be used")
        sys.exit()
    if options.writeFlows not in [None, "number", "probability"]:
        sys.stderr.write("Options --write-flows only accepts arguments 'number' and 'probability'")
        sys.exit()

    options.routeFiles = options.routeFiles.split(',')
    options.turnFiles = options.turnFiles.split(',') if options.turnFiles is not None else []
    options.edgeDataFiles = options.edgeDataFiles.split(',') if options.edgeDataFiles is not None else []
    options.odFiles = options.odFiles.split(',') if options.odFiles is not None else []
    if options.vehattrs and options.vehattrs[0] != ' ':
        options.vehattrs = ' ' + options.vehattrs

    if options.optimize is not None:
        try:
            import scipy.optimize  # noqa
            if options.optimize != "full":
                try:
                    options.optimize = int(options.optimize)
                except Exception:
                    print("Option optimize requires the value 'full' or an integer", file=sys.stderr)
                    sys.exit(1)
        except ImportError:
            print("Cannot use optimization (scipy not installed)", file=sys.stderr)
            sys.exit(1)

    if options.optimizeInput and type(options.optimize) != int:
        print("Option --optimize-input requires an integer argument for --optimize", file=sys.stderr)
        sys.exit(1)

    if options.threads > 1 and sys.version_info[0] < 3:
        print("Using multiple cpus is only supported for python 3", file=sys.stderr)
        sys.exit(1)

    return options


class CountData:
    def __init__(self, count, edgeTuple, allRoutes, isOD, options):
        self.origCount = count
        self.count = count
        self.edgeTuple = edgeTuple
        self.isOD = isOD
        self.options = options  # multiprocessing had issue with sumolib.options.getOptions().turnMaxGap
        self.routeSet = set()
        for routeIndex, edges in enumerate(allRoutes.unique):
            if self.routePasses(edges,):
                self.routeSet.add(routeIndex)

    def routePasses(self, edges):

        if self.isOD:
            return edges[0] == self.edgeTuple[0] and edges[-1] == self.edgeTuple[-1]
        try:
            i = edges.index(self.edgeTuple[0])
            maxDelta = self.options.turnMaxGap + 1
            for edge in self.edgeTuple[1:]:
                i2 = edges.index(edge, i)
                if i2 - i > maxDelta:
                    return False
                i = i2
        except ValueError:
            # first edge not in route
            return False
        return True

    def addCount(self, count):
        self.count += count
        self.origCount += count

    def sampleOpen(self, rng, openRoutes, routeCounts):
        cands = list(self.routeSet.intersection(openRoutes))
        assert (cands)
        probs = [routeCounts[i] for i in cands]
        x = rng.random() * sum(probs)
        seen = 0
        for route, prob in zip(cands, probs):
            seen += prob
            if seen >= x:
                return route
        assert (False)


def getIntervals(options):
    begin, end, interval = parseTimeRange(options.turnFiles + options.edgeDataFiles + options.odFiles)
    if options.begin is not None:
        begin = parseTime(options.begin)
    if options.end is not None:
        end = parseTime(options.end)
    if options.interval is not None:
        interval = parseTime(options.interval)

    result = []
    while begin < end:
        result.append((begin, begin + interval))
        begin += interval

    return result


def getOverlap(begin, end, iBegin, iEnd):
    """return overlap of the given intervals as fraction"""
    if iEnd <= begin or end <= iBegin:
        return 0  # no overlap
    elif iBegin >= begin and iEnd <= end:
        return 1  # data interval fully within requested interval
    elif iBegin <= begin and iEnd >= end:
        return (end - begin) / (iEnd - iBegin)  # only part of the data interval applies to the requested interval
    elif iBegin <= begin and iEnd <= end:
        return (iEnd - begin) / (iEnd - iBegin)  # partial overlap
    else:
        return (end - iBegin) / (iEnd - iBegin)  # partial overlap


def parseTurnCounts(interval, attr):
    for edgeRel in interval.edgeRelation:
        via = [] if edgeRel.via is None else edgeRel.via.split(' ')
        edges = tuple([edgeRel.attr_from] + via + [edgeRel.to])
        value = getattr(edgeRel, attr)
        yield edges, value


def parseEdgeCounts(interval, attr):
    for edge in interval.edge:
        yield (edge.id,), getattr(edge, attr)


def parseDataIntervals(parseFun, fnames, begin, end, allRoutes, attr, options, isOD=False, warn=False):
    locations = {}  # edges -> CountData
    result = []
    for fname in fnames:
        for interval in sumolib.xml.parse(fname, 'interval', heterogeneous=True):
            overlap = getOverlap(begin, end, parseTime(interval.begin), parseTime(interval.end))
            if overlap > 0:
                # print(begin, end, interval.begin, interval.end, "overlap:", overlap)
                for edges, value in parseFun(interval, attr):
                    try:
                        value = float(value)
                    except TypeError:
                        if warn:
                            print("Warning: Missing '%s' value in file '%s' for edge(s) '%s'" %
                                  (attr, fname, ' '.join(edges)), file=sys.stderr)
                        continue
                    if edges not in locations:
                        result.append(CountData(0, edges, allRoutes, isOD, options))
                        locations[edges] = result[-1]
                    elif isOD != locations[edges].isOD:
                        print("Warning: Edge relation '%s' occurs as turn relation and also as OD-relation" %
                              ' '.join(edges), file=sys.stderr)
                    locations[edges].addCount(int(value * overlap))
    return result


def parseTimeRange(fnames):
    begin = 1e20
    end = 0
    minInterval = 1e20
    for fname in fnames:
        for interval in sumolib.xml.parse(fname, 'interval'):
            iBegin = parseTime(interval.begin)
            iEnd = parseTime(interval.end)
            begin = min(begin, iBegin)
            end = max(end, iEnd)
            minInterval = min(minInterval, iEnd - iBegin)
    return begin, end, minInterval


def hasCapacity(dataIndices, countData):
    for i in dataIndices:
        if countData[i].count == 0:
            return False
    return True


def updateOpenRoutes(openRoutes, routeUsage, countData):
    return set(filter(lambda r: hasCapacity(routeUsage[r], countData), openRoutes))


def updateOpenCounts(openCounts, countData, openRoutes):
    return set(filter(lambda i: countData[i].routeSet.intersection(openRoutes), openCounts))


def optimize(options, countData, routes, usedRoutes, routeUsage):
    """ use relaxtion of the ILP problem for picking the number of times that each route is used
    x = usageCount vector (count for each route index)
    c = weight vector (vector of 1s)
    A_eq = routeUsage encoding
    b_eq = counts

    Rationale:
      c: costs for using each route,
         when minimizing x @ c, routes that pass multiple counting stations are getting an advantage

    """
    import scipy.optimize as opt
    import scipy.version
    import numpy as np

    m = len(countData)

    priorRouteCounts = getRouteCounts(routes, usedRoutes)

    relevantRoutes = [i for i in range(routes.number) if routeUsage[i]]
    priorRelevantRouteCounts = [priorRouteCounts[r] for r in relevantRoutes]
    k = len(relevantRoutes)

    if options.optimize == "full":
        # allow changing all prior usedRoutes
        bounds = None
    else:
        u = int(options.optimize)
        if u == 0:
            print("Optimization skipped")
            return
        # limited optimization: change prior routeCounts by at most u per route
        bounds = [(max(0, p - u), p + u) for p in priorRelevantRouteCounts] + [(0, None)] * m

    # Ax <= b
    # x + s = b
    # min s
    # -> x2 = [x, s]

    A = np.zeros((m, k))
    for i in range(0, m):
        for j in range(0, k):
            A[i][j] = int(relevantRoutes[j] in countData[i].routeSet)
    A_eq = np.concatenate((A, np.identity(m)), 1)

    # constraint: achieve counts
    b = np.asarray([cd.origCount for cd in countData])

    # minimization objective [routeCounts] + [slack]
    c = [options.minimizeVehs] * k + [1] * m

    # set x to prior counts and slack to deficit (otherwise solver may fail to find any soluton
    x0 = priorRelevantRouteCounts + [cd.origCount - cd.count for cd in countData]

    # print("k=%s" % k)
    # print("m=%s" % m)
    # print("A_eq (%s) %s" % (A_eq.shape, A_eq))
    # print("b (%s) %s" % (len(b), b))
    # print("c (%s) %s" % (len(c), c))
    # print("bounds (%s) %s" % (len(bounds) if bounds is not None else "-", bounds))
    # print("x0 (%s) %s" % (len(x0), x0))

    linProgOpts = {}
    if options.verbose:
        linProgOpts["disp"] = True

    try:
        res = opt.linprog(c, A_eq=A_eq, b_eq=b, bounds=bounds, x0=x0, options=linProgOpts)
    except TypeError:
        if options.verbose:
            print("Warning: Scipy version %s does not support initial guess for opt.linprog. Optimization may fail"
                  % scipy.version.version, file=sys.stderr)
        res = opt.linprog(c, A_eq=A_eq, b_eq=b, bounds=bounds, options=linProgOpts)

    if res.success:
        print("Optimization succeeded")
        del usedRoutes[:]
        routeCountsR = res.x[:k]  # cut of slack variables
        # translate to original route indices
        routeCounts = [0] * routes.number
        for index, count in zip(relevantRoutes, routeCountsR):
            routeCounts[index] = count

        # print("priorRouteCounts", priorRouteCounts)
        # print("relevantRoutes", relevantRoutes)
        # print("priorRelevantRouteCounts", priorRelevantRouteCounts)
        # print("routeCountsR", routeCountsR)
        # print("routeCounts", routeCounts)
        # slack = res.x[k:]
        # print("routeCounts (n=%s, sum=%s, intSum=%s, roundSum=%s) %s" % (
        #    len(routeCounts),
        #    sum(routeCounts),
        #    sum(map(int, routeCounts)),
        #    sum(map(round, routeCounts)),
        #    routeCounts))
        # print("slack (n=%s, sum=%s) %s" % (len(slack), sum(slack), slack))
        usedRoutes.extend(sum([[i] * int(round(c)) for i, c in enumerate(routeCounts)], []))
        # print("#usedRoutes=%s" % len(usedRoutes))
        # update countData
    else:
        print("Optimization failed")


def zero():
    # cannot pickle lambda for multiprocessing
    return 0


class Routes:
    def __init__(self, routefiles):
        self.all = []
        self.edgeProbs = defaultdict(zero)
        self.edgeIDs = {}
        self.withProb = 0
        for routefile in routefiles:
            # not all routes may have specified probability, in this case use their number of occurence
            for r in sumolib.xml.parse(routefile, 'route', heterogeneous=True):
                edges = tuple(r.edges.split())
                self.all.append(edges)
                prob = float(r.getAttributeSecure("probability", 1))
                if r.hasAttribute("probability"):
                    self.withProb += 1
                    prob = float(r.probability)
                else:
                    prob = 1
                if prob <= 0:
                    print("Warning: route probability must be positive (edges=%s)" % r.edges, file=sys.stderr)
                    prob = 0
                if r.hasAttribute("id"):
                    self.edgeIDs[edges] = r.id
                self.edgeProbs[edges] += prob
        self.unique = sorted(list(self.edgeProbs.keys()))
        self.number = len(self.unique)
        self.edges2index = dict([(e, i) for i, e in enumerate(self.unique)])
        if len(self.unique) == 0:
            print("Error: no input routes loaded", file=sys.stderr)
            sys.exit()
        self.probabilities = np.array([self.edgeProbs[e] for e in self.unique], dtype=np.float64)


def resetCounts(usedRoutes, routeUsage, countData):
    for cd in countData:
        cd.count = cd.origCount
    for r in usedRoutes:
        for i in routeUsage[r]:
            countData[i].count -= 1


def getRouteCounts(routes, usedRoutes):
    result = [0] * routes.number
    for r in usedRoutes:
        result[r] += 1
    return result


def getRouteUsage(routes, countData):
    # store which counting locations are used by each route (using countData index)
    routeUsage = [set() for r in routes.unique]
    for i, cd in enumerate(countData):
        for routeIndex in cd.routeSet:
            routeUsage[routeIndex].add(i)
    return routeUsage


def main(options):
    if options.seed:
        random.seed(options.seed)
        np.random.seed(options.seed)

    routes = Routes(options.routeFiles)

    intervals = getIntervals(options)

    # preliminary integrity check for the whole time range
    b = intervals[0][0]
    e = intervals[-1][-1]
    countData = (parseDataIntervals(parseTurnCounts, options.turnFiles, b, e,
                                    routes, options.turnAttr, options=options, warn=True) +
                 parseDataIntervals(parseEdgeCounts, options.edgeDataFiles, b, e,
                                    routes, options.edgeDataAttr, options=options, warn=True) +
                 parseDataIntervals(parseTurnCounts, options.odFiles, b, e,
                                    routes, options.turnAttr, options=options, isOD=True, warn=True))
    routeUsage = getRouteUsage(routes, countData)

    for cd in countData:
        if cd.count > 0 and not cd.routeSet:
            print("Warning: no routes pass edge '%s' (count %s)" % (
                ' '.join(cd.edgeTuple), cd.count), file=sys.stderr)

    if options.verbose:
        print("Loaded %s routes (%s distinct)" % (len(routes.all), routes.number))
        if options.weighted:
            print("Loaded probability for %s routes" % routes.withProb)
        if options.verboseHistogram:
            edgeCount = sumolib.miscutils.Statistics("route edge count", histogram=True)
            detectorCount = sumolib.miscutils.Statistics("route detector count", histogram=True)
            for i, edges in enumerate(routes.unique):
                edgeCount.add(len(edges), i)
                detectorCount.add(len(routeUsage[i]), i)
            print("input %s" % edgeCount)
            print("input %s" % detectorCount)

    mismatchf = None
    if options.mismatchOut:
        mismatchf = open(options.mismatchOut, 'w')
        sumolib.writeXMLHeader(mismatchf, "$Id$")  # noqa
        mismatchf.write('<data>\n')

    underflowSummary = sumolib.miscutils.Statistics("all interval underflow")
    overflowSummary = sumolib.miscutils.Statistics("all interval overflow")
    gehSummary = sumolib.miscutils.Statistics("all interval GEH%")

    with open(options.out, 'w') as outf:
        sumolib.writeXMLHeader(outf, "$Id$", "routes")  # noqa
        if options.threads > 1:
            # call the multiprocessing function
            results = multi_process(options.threads, options.seed, intervals,
                                    _solveIntervalMP, outf, mismatchf, options=options, routes=routes)
            # handle the uFlow, oFlow and GEH
            for _, result in results:
                for i, begin in enumerate(result[0]):
                    underflowSummary.add(result[1][i], begin)
                    overflowSummary.add(result[2][i], begin)
                    gehSummary.add(result[3][i], begin)
        else:
            for begin, end in intervals:
                intervalPrefix = "" if len(intervals) == 1 else "%s_" % int(begin)
                uFlow, oFlow, gehOK, _ = solveInterval(options, routes, begin, end, intervalPrefix, outf, mismatchf)
                underflowSummary.add(uFlow, begin)
                overflowSummary.add(oFlow, begin)
                gehSummary.add(gehOK, begin)
        outf.write('</routes>\n')

    if options.mismatchOut:
        mismatchf.write('</data>\n')
        mismatchf.close()

    if len(intervals) > 1:
        print(underflowSummary)
        print(overflowSummary)
        print(gehSummary)


def _sample(sampleSet, rng):
    population = tuple(sampleSet)
    return population[(int)(rng.random() * len(population))]


def _sample_skewed(sampleSet, rng, probabilityMap):
    # build cumulative distribution function for weighted sampling
    cdf = []
    total = 0
    population = tuple(sampleSet)
    for element in population:
        total += probabilityMap[element]
        cdf.append(total)

    value = random.random() * total
    return population[np.searchsorted(cdf, value)]


def _solveIntervalMP(options, routes, interval, cpuIndex):
    output_list = []
    rng = random.Random()
    rng.seed(options.seed + cpuIndex)
    for begin, end in interval:
        local_outf = StringIO()
        local_mismatch_outf = StringIO() if options.mismatchOut else None
        intervalPrefix = "%s_" % int(begin)
        uFlow, oFlow, gehOKNum, local_outf = solveInterval(
            options, routes, begin, end, intervalPrefix, local_outf, local_mismatch_outf, rng=rng)
        output_list.append([begin, uFlow, oFlow, gehOKNum, local_outf.getvalue(),
                            local_mismatch_outf.getvalue() if options.mismatchOut else None])
    output_lst = list(zip(*output_list))
    return output_lst


def solveInterval(options, routes, begin, end, intervalPrefix, outf, mismatchf, rng=random):
    # store which routes are passing each counting location (using route index)
    countData = (parseDataIntervals(parseTurnCounts, options.turnFiles, begin, end, routes, options.turnAttr,
                                    options=options)
                 + parseDataIntervals(parseEdgeCounts, options.edgeDataFiles, begin, end, routes,
                                      options.edgeDataAttr,
                                      options=options)
                 + parseDataIntervals(parseTurnCounts, options.odFiles, begin, end, routes, options.turnAttr,
                                      isOD=True,
                                      options=options)
                 )

    routeUsage = getRouteUsage(routes, countData)
    unrestricted = set([r for r, usage in enumerate(routeUsage) if len(usage) == 0])
    if options.verbose and len(unrestricted) > 0:
        print("Ignored %s routes which do not pass any counting location" % len(unrestricted))

    # pick a random counting location and select a new route that passes it until
    # all counts are satisfied or no routes can be used anymore
    openRoutes = set(range(0, routes.number))
    openCounts = set(range(0, len(countData)))
    openRoutes = updateOpenRoutes(openRoutes, routeUsage, countData)
    openCounts = updateOpenCounts(openCounts, countData, openRoutes)
    openRoutes = openRoutes.difference(unrestricted)

    usedRoutes = []
    if options.optimizeInput:
        usedRoutes = [routes.edges2index[e] for e in routes.all]
        resetCounts(usedRoutes, routeUsage, countData)
    else:
        while openCounts:
            if options.weighted:
                routeIndex = _sample_skewed(openRoutes, rng, routes.probabilities)
            else:
                # sampling equally among open counting locations appears to
                # improve GEH but it would also introduce a bias in the loaded
                # route probabilities
                cd = countData[_sample(openCounts, rng)]
                routeIndex = _sample(cd.routeSet.intersection(openRoutes), rng)
            usedRoutes.append(routeIndex)
            for dataIndex in routeUsage[routeIndex]:
                countData[dataIndex].count -= 1
            openRoutes = updateOpenRoutes(openRoutes, routeUsage, countData)
            openCounts = updateOpenCounts(openCounts, countData, openRoutes)

    totalMismatch = sum([cd.count for cd in countData])
    if totalMismatch > 0 and options.optimize is not None:
        if options.verbose:
            print("Starting optimization for interval [%s, %s] (mismatch %s)" % (
                begin, end, totalMismatch))
        optimize(options, countData, routes, usedRoutes, routeUsage)
        resetCounts(usedRoutes, routeUsage, countData)
    # avoid bias from sampling order / optimization
    random.shuffle(usedRoutes, rng.random)

    if usedRoutes:
        outf.write('<!-- begin="%s" end="%s" -->\n' % (begin, end))
        period = (end - begin) / len(usedRoutes)
        depart = begin
        routeCounts = getRouteCounts(routes, usedRoutes)
        if options.writeRouteIDs:
            for routeIndex in sorted(set(usedRoutes)):
                edges = routes.unique[routeIndex]
                routeIDComment = ""
                if edges in routes.edgeIDs:
                    routeIDComment = " (%s)" % routes.edgeIDs[edges]
                outf.write('    <route id="%s%s" edges="%s"/> <!-- %s%s -->\n' % (
                    intervalPrefix, routeIndex, ' '.join(edges),
                    routeCounts[routeIndex], routeIDComment))
            outf.write('\n')
        elif options.writeRouteDist:
            outf.write('    <routeDistribution id="%s%s"/>\n' % (intervalPrefix, options.writeRouteDist))
            for routeIndex in sorted(set(usedRoutes)):
                outf.write('        <route id="%s%s" edges="%s" probability="%s"/>\n' % (
                    intervalPrefix, routeIndex, ' '.join(routes.unique[routeIndex]), routeCounts[routeIndex]))
            outf.write('    </routeDistribution>\n\n')

        routeID = options.writeRouteDist
        if options.writeFlows is None:
            for i, routeIndex in enumerate(usedRoutes):
                if options.writeRouteIDs:
                    routeID = routeIndex
                vehID = options.prefix + intervalPrefix + str(i)
                if routeID is not None:
                    outf.write('    <vehicle id="%s" depart="%.2f" route="%s%s"%s/>\n' % (
                        vehID, depart, intervalPrefix, routeID, options.vehattrs))
                else:
                    outf.write('    <vehicle id="%s" depart="%.2f"%s>\n' % (
                        vehID, depart, options.vehattrs))
                    outf.write('        <route edges="%s"/>\n' % ' '.join(routes.unique[routeIndex]))
                    outf.write('    </vehicle>\n')
                depart += period
        else:
            routeDeparts = defaultdict(list)
            for routeIndex in usedRoutes:
                routeDeparts[routeIndex].append(depart)
                depart += period
            if options.writeRouteDist:
                totalCount = sum(routeCounts)
                probability = totalCount / (end - begin)
                flowID = options.prefix + intervalPrefix + options.writeRouteDist
                if options.writeFlows == "number" or probability > 1.001:
                    repeat = 'number="%s"' % totalCount
                    if options.writeFlows == "probability":
                        sys.stderr.write("Warning: could not write flow %s with probability %.2f\n" %
                                         (flowID, probability))
                else:
                    repeat = 'probability="%s"' % probability
                outf.write('    <flow id="%s" begin="%.2f" end="%.2f" %s route="%s"%s/>\n' % (
                    flowID, begin, end, repeat,
                    options.writeRouteDist, options.vehattrs))
            else:
                # ensure flows are sorted
                flows = []
                for routeIndex in sorted(set(usedRoutes)):
                    outf2 = StringIO()
                    fBegin = min(routeDeparts[routeIndex])
                    fEnd = max(routeDeparts[routeIndex] + [fBegin + 1.0])
                    probability = routeCounts[routeIndex] / (fEnd - fBegin)
                    flowID = "%s%s%s" % (options.prefix, intervalPrefix, routeIndex)
                    if options.writeFlows == "number" or probability > 1.001:
                        repeat = 'number="%s"' % routeCounts[routeIndex]
                        if options.writeFlows == "probability":
                            sys.stderr.write("Warning: could not write flow %s with probability %.2f\n" % (
                                flowID, probability))
                    else:
                        repeat = 'probability="%s"' % probability
                    if options.writeRouteIDs:
                        outf2.write('    <flow id="%s" begin="%.2f" end="%.2f" %s route="%s%s"%s/>\n' % (
                            flowID, fBegin, fEnd, repeat,
                            intervalPrefix, routeIndex, options.vehattrs))
                    else:
                        outf2.write('    <flow id="%s" begin="%.2f" end="%.2f" %s%s>\n' % (
                            flowID, fBegin, fEnd, repeat, options.vehattrs))
                        outf2.write('        <route edges="%s"/>\n' % ' '.join(routes.unique[routeIndex]))
                        outf2.write('    </flow>\n')
                    flows.append((fBegin, outf2))
                flows.sort()
                for fBegin, outf2 in flows:
                    outf.write(outf2.getvalue())

    underflow = sumolib.miscutils.Statistics("underflow locations")
    overflow = sumolib.miscutils.Statistics("overflow locations")
    gehStats = sumolib.miscutils.Statistics("GEH")
    numGehOK = 0.0
    hourFraction = (end - begin) / 3600.0
    totalCount = 0
    totalOrigCount = 0
    for cd in countData:
        localCount = cd.origCount - cd.count
        totalCount += localCount
        totalOrigCount += cd.origCount
        if cd.count > 0:
            underflow.add(cd.count, cd.edgeTuple)
        elif cd.count < 0:
            overflow.add(cd.count, cd.edgeTuple)
        origHourly = cd.origCount / hourFraction
        localHourly = localCount / hourFraction
        geh = sumolib.miscutils.geh(origHourly, localHourly)
        if geh < options.gehOk:
            numGehOK += 1
        gehStats.add(geh, "[%s] %s %s" % (
            ' '.join(cd.edgeTuple), int(origHourly), int(localHourly)))

    outputIntervalPrefix = "" if intervalPrefix == "" else "%s: " % int(begin)
    countPercentage = gehOK = "%.2f%%" % (100 * totalCount / float(totalOrigCount)) if totalOrigCount else "-"
    gehOKNum = 100 * numGehOK / float(len(countData)) if countData else 100
    gehOK = "%.2f%%" % gehOKNum if countData else "-"
    print("%sWrote %s routes (%s distinct) achieving total count %s (%s) at %s locations. GEH<%s for %s" % (
        outputIntervalPrefix,
        len(usedRoutes), len(set(usedRoutes)),
        totalCount, countPercentage, len(countData),
        options.gehOk, gehOK))

    if options.verboseHistogram:
        edgeCount = sumolib.miscutils.Statistics("route edge count", histogram=True)
        detectorCount = sumolib.miscutils.Statistics("route detector count", histogram=True)
        for i, r in enumerate(usedRoutes):
            edgeCount.add(len(routes.unique[r]), i)
            detectorCount.add(len(routeUsage[r]), i)
        print("result %s" % edgeCount)
        print("result %s" % detectorCount)
        print(gehStats)

    if underflow.count() > 0:
        print("Warning: %s (total %s)" % (underflow, sum(underflow.values)))
    if overflow.count() > 0:
        print("Warning: %s (total %s)" % (overflow, sum(overflow.values)))
    sys.stdout.flush()  # needed for multiprocessing

    if mismatchf:
        mismatchf.write('    <interval id="deficit" begin="%s" end="%s">\n' % (begin, end))
        for cd in countData:
            if len(cd.edgeTuple) == 1:
                mismatchf.write('        <edge id="%s" measuredCount="%s" deficit="%s"/>\n' % (
                    cd.edgeTuple[0], cd.origCount, cd.count))
            elif len(cd.edgeTuple) == 2:
                mismatchf.write('        <edgeRelation from="%s" to="%s" measuredCount="%s" deficit="%s"/>\n' % (
                    cd.edgeTuple[0], cd.edgeTuple[1], cd.origCount, cd.count))
            else:
                print("Warning: output for edge relations with more than 2 edges not supported (%s)" % cd.edgeTuple,
                      file=sys.stderr)
        mismatchf.write('    </interval>\n')

    return sum(underflow.values), sum(overflow.values), gehOKNum, outf


if __name__ == "__main__":
    main(get_options())



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    jtcrouter.py
# @author  Jakob Erdmann
# @date    2019-11-07

"""
Generate routes like jtrrouter but bas on count parameters at connections
The counts are used to derive the turn file and flow file for calling jtrrouter
"""
from __future__ import absolute_import
from __future__ import print_function

import os
import sys
from argparse import ArgumentParser
from collections import defaultdict
import subprocess

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools/turn-defs'))
import sumolib  # noqa
from turnCount2EdgeCount import parseEdgeCounts  # noqa


def get_options(args=None):
    parser = ArgumentParser(description="Route by turn counts")
    parser.add_argument("-n", "--net-file", dest="net", help="Input net file")
    parser.add_argument("-t", "--turn-file", dest="turnFile", help="Input turn-count file")
    parser.add_argument("-o", "--output-file", dest="out", default="out.rou.xml",
                        help="Output route file")
    parser.add_argument("--turn-output", dest="turnOutput", default="turns.tmp.xml",
                        help="Intermediate turn-ratio-file")
    parser.add_argument("--flow-output", dest="flowOuput", default="flows.tmp.xml",
                        help="Intermediate flow file")
    parser.add_argument("--turn-attribute", dest="turnAttr", default="count",
                        help="Read turning counts from the given attribute")
    parser.add_argument("-b", "--begin",  default=0, help="begin time")
    parser.add_argument("-e", "--end",  default=3600, help="end time (default 3600)")
    parser.add_argument("-p", "--count-param", dest="countParam", default="count",
                        help="the connection parameter to use as count")
    parser.add_argument("--fringe-flows", action="store_true", default=False, dest="fringe_flows",
                        help="Avoid overlapping flows (start only on the outside of the network)")
    parser.add_argument("--discount-sources", "-D",  action="store_true", default=False, dest="discountSources",
                        help="passes option --discount-sources to jtrrouter")
    parser.add_argument("--prefix", dest="prefix", default="",
                        help="prefix for the flow ids")
    parser.add_argument("-a", "--attributes", dest="flowattrs", default="",
                        help="additional flow attributes")
    options = parser.parse_args(args=args)
    if options.net is None:
        parser.print_help()
        sys.exit()
    if options.flowattrs and options.flowattrs[0] != ' ':
        options.flowattrs = ' ' + options.flowattrs
    return options


def findFringe(edge, countParam, intermediateCounts=None):
    if intermediateCounts:
        # do not backtrack past edges that define turning counts (to avoid duplicate flows)
        return None
    if edge.is_fringe(edge._incoming):
        return edge
    elif len(edge.getIncoming()) == 1:
        prev = edge.getIncoming().keys()[0]
        return findFringe(prev, countParam, getCounts(prev, countParam))
    return None


def getCounts(edge, countParam):
    counts = defaultdict(lambda: 0)
    for toEdge, cons in edge.getOutgoing().items():
        for con in cons:
            value = con.getParam(countParam)
            if value is not None:
                counts[con.getTo().getID()] += float(value)
    return counts


def main(options):
    if options.turnFile is None:
        # read data from connection params
        net = sumolib.net.readNet(options.net)
        with open(options.turnOutput, 'w') as tf, open(options.flowOuput, 'w') as ff:
            sumolib.writeXMLHeader(tf, "$Id$", "turns")  # noqa
            sumolib.writeXMLHeader(ff, "$Id$", "routes")  # noqa
            tf.write('    <interval begin="%s" end="%s">\n' % (options.begin, options.end))
            for edge in net.getEdges():
                counts = getCounts(edge, options.countParam)
                if counts:
                    tf.write('        <fromEdge id="%s">\n' % (edge.getID()))
                    for toEdge, count in counts.items():
                        tf.write('            <toEdge id="%s" probability="%s"/>\n' % (toEdge, count))
                    tf.write('        </fromEdge>\n')

                    totalCount = int(sum(counts.values()))
                    fromEdge = edge
                    if options.fringe_flows:
                        fromEdge = findFringe(edge, options.countParam)
                    if fromEdge:
                        ff.write('    <flow id="%s%s" from="%s" begin="%s" end="%s" number="%s"%s/>\n' % (
                            options.prefix, edge.getID(), fromEdge.getID(),
                            options.begin, options.end,
                            totalCount, options.flowattrs))
            tf.write('    </interval>\n')
            tf.write('</turns>\n')
            ff.write('</routes>\n')
    else:
        # read turn-count file
        options.turnOutput = options.turnFile
        with open(options.flowOuput, 'w') as ff:
            ff.write('<routes>\n')
            for i, interval in enumerate(parseEdgeCounts(options.turnFile, options.turnAttr)):
                interval_id, interval_begin, interval_end, counts = interval
                for edge in sorted(counts.keys()):
                    count = counts[edge]
                    if count > 0:
                        flowID = edge
                        if i > 0:
                            flowID += "#%s" % i
                        ff.write('    <flow id="%s%s" from="%s" begin="%s" end="%s" number="%s"%s/>\n' % (
                            options.prefix,
                            flowID, edge, interval_begin, interval_end,
                            int(count),
                            options.flowattrs))
            ff.write('</routes>\n')

    JTRROUTER = sumolib.checkBinary('jtrrouter')
    args = [JTRROUTER,
            '-n', options.net,
            '--turn-ratio-files', options.turnOutput,
            '--route-files', options.flowOuput,
            '--accept-all-destinations',
            '-o', options.out]
    if not options.fringe_flows:
        args += ['-S']
        if options.discountSources:
            args += ['-D']
    subprocess.call(args)


if __name__ == "__main__":
    main(get_options())



#!/usr/bin/env python3

""" SUMO Activity-Based Mobility Generator

    Author: Lara CODECA

    This program and the accompanying materials are made available under the
    terms of the Eclipse Public License 2.0 which is available at
    http://www.eclipse.org/legal/epl-2.0.
"""

import argparse
import collections
import cProfile

import io
import json
import logging
import os
from pprint import pformat
import pstats
import sys
import xml.etree.ElementTree

from enum import Enum

import numpy
from numpy.random import RandomState
from tqdm import tqdm

from agsrc import activities, environment, sagaexceptions, sumoutils

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    import traci
    import traci.constants as tc
    from traci._simulation import Stage
else:
    sys.exit("Please declare environment variable 'SUMO_HOME'")

def get_options(cmd_args):
    """ Argument Parser. """
    parser = argparse.ArgumentParser(
        prog='activitygen.py', usage='%(prog)s -c configuration.json',
        description='SUMO Activity-Based Mobility Generator')
    parser.add_argument(
        '-c', type=str, dest='config', required=True,
        help='JSON configuration file.')
    parser.add_argument(
        '--profiling', dest='profiling', action='store_true',
        help='Enable Python3 cProfile feature.')
    parser.add_argument(
        '--no-profiling', dest='profiling', action='store_false',
        help='Disable Python3 cProfile feature.')
    parser.set_defaults(profiling=False)
    return parser.parse_args(cmd_args)

class MobilityGenerator():
    """ Generates intermodal mobility for SUMO starting from a synthetic population. """

    class ModeShare(Enum):
        """
        Selector between two interpretation of the values used for the modes:
            - PROBABILITY: only one mode is selected using the given probability.
            - WEIGHT: all the modes are generated, the cost is multiplied by the given weight and
                        only the cheapest solution is used.
        """
        PROBABILITY = 1
        WEIGHT = 2

    LAST_STOP_PLACEHOLDER = -42.42

    def _configure_loggers(self):
        """ Setup the console and file logger. """
        self.logger = logging.getLogger('ActivityGen')
        self.logger.setLevel(logging.DEBUG)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(logging.INFO)
        _console_handler.setFormatter(
            logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s'))
        self.logger.addHandler(_console_handler)
        _file_handler = logging.FileHandler('{}.debug.log'.format(self._conf['outputPrefix']))
        _file_handler.setLevel(logging.DEBUG)
        _file_handler.setFormatter(
            logging.Formatter(
                '[%(asctime)s]:[%(name)s]:[%(module)s:%(funcName)s:%(lineno)d]:[%(levelname)s] '
                '%(message)s'))
        self.logger.addHandler(_file_handler)

    def __init__(self, conf, profiling=False):
        """
        Initialize the synthetic population.
            :param conf: distionary with the configurations
            :param profiling=False: enable cProfile
        """
        self._conf = conf

        # Setting the LOGGING
        self._configure_loggers()

        if not 'modeSelection' in conf['intermodalOptions']:
            raise Exception('The parameter "modeSelection" in "intermodalOptions" must be defined.')
        self._mode_interpr = None
        if conf['intermodalOptions']['modeSelection'] == 'PROBABILITY':
            self._mode_interpr = self.ModeShare.PROBABILITY
        elif conf['intermodalOptions']['modeSelection'] == 'WEIGHT':
            self._mode_interpr = self.ModeShare.WEIGHT
        else:
            raise Exception('The parameter "modeSelection" in "intermodalOptions" must be set to '
                            '"PROBABILITY" or "WEIGHT".')

        self._max_retry_number = 1000
        if 'maxNumTry' in conf:
            self._max_retry_number = conf['maxNumTry']

        self._profiling = profiling

        self._random_generator = RandomState(seed=self._conf['seed'])

        self.logger.info('Starting TraCI with file %s.', conf['sumocfg'])
        traci.start(['sumo', '-c', conf['sumocfg']], traceFile='traci.log')

        if not 'vTypes' in self._conf['SUMOadditionals']:
            raise Exception('The parameter "vTypes" in "SUMOadditionals" must be defined.')
        self._vtype_to_vclasses = collections.defaultdict(list)
        self._load_vtypes_distributions(self._conf['SUMOadditionals']['vTypes'])

        self._env = environment.Environment(
            conf, sumo=traci, profiling=profiling, logger=self.logger)
        self._chains = activities.Activities(
            conf, sumo=traci, environment=self._env, profiling=profiling, logger=self.logger)

        self._all_trips = collections.defaultdict(dict)

        self.logger.info('Computing the number of entities for each mobility slice..')
        self._compute_entities_per_slice()

    def generate(self):
        """ Generates and saves the mobility. """
        self._mobility_generation()
        self._save_mobility()
        self._close_traci()

    def _mobility_generation(self):
        """ Generate the mobility for the synthetic population. """
        self.logger.info('Generating on-deman fleet..')
        self._generate_taxi_fleet()
        self.logger.info('Generating trips for each mobility slice..')
        self._compute_trips_per_slice()

    def _save_mobility(self):
        """ Save the generated trips to files. """
        self.logger.info('Saving trips files..')
        if self._conf['mergeRoutesFiles']:
            self._saving_trips_to_single_file()
        else:
            self._saving_trips_to_files()

    def _close_traci(self):
        """ Artefact to close TraCI properly. """
        self.logger.debug('Closing TraCI.')
        traci.close()

    ## ---------------------------------------------------------------------------------------- ##
    ##                                Loaders                                                   ##
    ## ---------------------------------------------------------------------------------------- ##

    def _load_vtypes_distributions(self, filename):
        """ Load vTypeDistributions from XML file. """
        xml_tree = xml.etree.ElementTree.parse(filename).getroot()
        for child in xml_tree:
            if child.tag == 'vType':
                self._vtype_to_vclasses[child.attrib['id']].append(child.attrib['vClass'])
            elif child.tag == 'vTypeDistribution':
                for subchild in child:
                    self._vtype_to_vclasses[child.attrib['id']].append(subchild.attrib['vClass'])
        self.logger.debug('vTypes loaded: \n%s', pformat(self._vtype_to_vclasses))

    ## ---------------------------------------------------------------------------------------- ##
    ##                                Mobility Generation                                       ##
    ## ---------------------------------------------------------------------------------------- ##

    ## ----                       On-demand fleet generation                               ---- ##

    ON_DEMAND_STR = """
    <vehicle id="{id}" type="on-demand" depart="0.0">
        <route edges="{edge}"/>
        <stop lane="{lane}" startPos="1.0" endPos="-1.0" triggered="person"/>
    </vehicle>"""

    def _generate_taxi_fleet(self):
        """ Generate the number of on-demand vehicles set in the configuration file. """
        self.logger.info('On-demand fleet expected size of %d',
                     self._conf['intermodalOptions']['taxiFleetSize'])
        _fleet = []
        for _vehicle_id in tqdm(range(self._conf['intermodalOptions']['taxiFleetSize'])):
            _name = 'on-demand.{}'.format(_vehicle_id)
            _lane = self._env.get_random_lane_from_tazs()
            if _lane is None:
                continue
            _edge = _lane.split('_')[0]
            _fleet.append(
                {
                    'id': _name,
                    'depart': 0.0,
                    'string': self.ON_DEMAND_STR.format(id=_name, edge=_edge, lane=_lane)
                }
            )
        self.logger.info('Generated an on-demant fleet of %d vehicles.', len(_fleet))
        self._all_trips['on-demand-fleet'][0] = _fleet
        with open('{}.trips.dump.json'.format(self._conf['outputPrefix']), 'w') as openfile:
            json.dump(
                {
                    'name': 'on-demand-fleet',
                    'depart': 0,
                    'taxi': _fleet,
                }, openfile, indent=2)

    ## ----                           Person trips generation                               ---- ##

    def _compute_entities_per_slice(self):
        """
        Compute the absolute number of entities that are going to be created
        for each moblitiy slice, given a population.
        """
        self.logger.info('Population: %d', self._conf['population']['entities'])

        for m_slice in self._conf['slices'].keys():
            self._conf['slices'][m_slice]['tot'] = int(
                self._conf['population']['entities'] * self._conf['slices'][m_slice]['perc'])
            self.logger.info('\t %s: %d', m_slice, self._conf['slices'][m_slice]['tot'])

    def _compute_trips_per_slice(self):
        """ Compute the trips for the synthetic population for each mobility slice. """

        total = 0

        _modes_stats = collections.defaultdict(int)
        _chains_stats = collections.defaultdict(int)

        for name, m_slice in self._conf['slices'].items():
            self.logger.info('[%s] Computing %d trips from %s to %s ... ',
                         name, m_slice['tot'], m_slice['loc_origin'], m_slice['loc_primary'])

            ## Activity chains preparation
            activity_chains = []
            activity_chains_weights = []
            for _weight, _chain, _modes in m_slice['activityChains']:
                activity_chains.append((_chain, _modes))
                activity_chains_weights.append(_weight)
            activity_index = range(len(activity_chains))

            if self._profiling:
                _pr = cProfile.Profile()
                _pr.enable()

            for entity_id in tqdm(range(m_slice['tot'])):
                ## Select the activity chain
                _index = self._random_generator.choice(
                    activity_index, p=activity_chains_weights)
                _chain, _modes = activity_chains[_index]
                self.logger.debug('_compute_trips_per_slice: Chain: %s', '{}'.format(_chain))
                self.logger.debug('_compute_trips_per_slice: Modes: %s', '{}'.format(_modes))

                _person_trip = None

                # (Intermodal) trip
                _final_chain = None
                _stages = None
                _error_counter = 0
                while not _person_trip and _error_counter < self._max_retry_number:
                    try:
                        self.logger.debug(
                            ' ====================== _generate_trip ====================== ')
                        _final_chain, _stages, _selected_mode = self._generate_trip(
                            self._conf['taz'][m_slice['loc_origin']],
                            self._conf['taz'][m_slice['loc_primary']],
                            _chain, _modes)

                        ## Generating departure time
                        _depart = numpy.round(_final_chain[1].start, decimals=2)
                        if _depart < 0.0:
                            _msg_error = 'Negative departure time generated: {}.'.format(_depart)
                            _msg_error += ' Required fixing of the start time of the initial'
                            _msg_error += ' activities. '
                            # raise sagaexceptions.TripGenerationGenericError(_msg_error)
                            _depart = 0.0
                            _msg_error += ' Departure set to 0.0.'
                            self.logger.critical(_msg_error)

                        if _depart not in self._all_trips[name].keys():
                            self._all_trips[name][_depart] = []

                        _person_trip = {
                            'id': '{}_{}'.format(name, entity_id),
                            'depart': _depart,
                            'stages': _stages,
                        }

                        complete_trip = self._generate_sumo_trip_from_activitygen(_person_trip)

                        _person_trip['string'] = complete_trip
                        ## For statistical purposes.
                        _modes_stats[_selected_mode] += 1
                        _chains_stats[self._hash_final_chain(_final_chain)] += 1

                    except sagaexceptions.TripGenerationGenericError:
                        _person_trip = None
                        _error_counter += 1
                    finally:
                        self.logger.debug(' ====================== DONE ====================== ')

                if _person_trip:
                    # Trip creation
                    self._all_trips[name][_depart].append(_person_trip)
                    with open(
                        '{}.trips.dump.json'.format(self._conf['outputPrefix']), 'a') as openfile:
                        json.dump(
                            {
                                'name': name,
                                'depart': _depart,
                                'personTrip': pformat(_person_trip),
                            }, openfile, indent=2)
                    self.logger.debug('Generated: %s', _person_trip['string'])
                    total += 1

                else:
                    self.logger.critical(
                        '_generate_trip from %s to %s generated %d errors, '
                        'trip generation aborted..',
                        self._conf['taz'][m_slice['loc_origin']],
                        self._conf['taz'][m_slice['loc_primary']],
                        _error_counter)

            if self._profiling:
                _pr.disable()
                _s = io.StringIO()
                _ps = pstats.Stats(_pr, stream=_s).sort_stats('cumulative')
                _ps.print_stats(10)
                print(_s.getvalue())
                input("Press any key to continue..")

        self.logger.info('Generated %d trips.', total)
        self.logger.info('Mode splits:')
        for mode, value in _modes_stats.items():
            self.logger.info('\t %s: %d (%.2f).', mode, value, float(value/total))
        self.logger.info('Activity chains splits:')
        for chain, value in _chains_stats.items():
            self.logger.info('\t %s: %d (%.2f).', chain, value, float(value/total))

    @staticmethod
    def _hash_final_chain(chain):
        activity_list = list()
        for pos in range(1, len(chain)+1):
            activity_list.append(chain[pos].activity)
        return pformat(activity_list)

    ## ---- Functions for _compute_trips_per_slice:
    #                                       _generate_trip, _generate_intermodal_trip_traci ---- ##

    def _generate_trip(self, from_area, to_area, activity_chain, modes):
        """ Returns the trip for the given activity chain. """

        trip = None
        solutions = []

        ## Mode selection, weight or probability
        _interpr_modes = None
        if self._mode_interpr == self.ModeShare.PROBABILITY:
            _probs = []
            _vals = []
            for mode, prob in modes:
                _vals.append(mode)
                _probs.append(prob)
            selection = self._random_generator.choice(_vals, p=_probs)
            _interpr_modes = [[selection, 1.0]] ## Unique mode, without weight.
        else:
            _interpr_modes = modes

        for mode, weight in _interpr_modes:

            ### DEBUGGING MODES: # <==================== !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # mode = 'walk'
            # mode = 'bicycle'
            # mode = 'public'
            # mode = 'passenger'
            # mode = 'motorcycle'
            # mode = 'on-demand'
            # mode = 'commercial'
            # mode = 'special'
            # mode = 'evehicle'

            _person_steps = None
            _error_counter = 0
            while not _person_steps and _error_counter < self._max_retry_number:
                try:
                    _person_steps, _person_stages = self._generate_intermodal_trip_traci(
                        from_area, to_area, activity_chain, mode)
                except sagaexceptions.TripGenerationGenericError:
                    _person_steps = None
                    _error_counter += 1

            if _person_steps:
                ## Cost computation.
                solutions.append((sumoutils.cost_from_route(_person_steps) * weight,
                                  _person_steps, _person_stages, mode))
            else:
                self.logger.critical(
                    '_generate_intermodal_trip_traci from "%s" to "%s" with "%s" '
                    'generated %d errors, trip generation aborted..',
                    from_area, to_area, mode, _error_counter)

        ## Compose the final person trip.
        if solutions:
            ## For the moment, the best solution is the one with minor cost.
            best = sorted(solutions)[0] ## Ascending.
            trip = (best[2], best[1], best[3]) ## _person_stages, _person_steps, mode
        else:
            raise sagaexceptions.TripGenerationRouteError(
                'No solution foud for chain {} and modes {}.'.format(
                    activity_chain, _interpr_modes))
        return trip

    def _generate_intermodal_trip_traci(self, from_area, to_area, activity_chain, mode):
        """ Return the person trip for a given mode generated with TraCI """

        self.logger.debug(
            ' ====================== _generate_intermodal_trip_traci ====================== ')

        _person_stages = self._chains.generate_person_stages(
            from_area, to_area, activity_chain, mode)

        _person_steps = []
        _current_depart_time = None

        _mode, _ptype, _vtype = sumoutils.get_intermodal_mode_parameters(
            mode, self._conf['intermodalOptions']['vehicleAllowedParking'])
        self.logger.debug('Mode: %s ==> %s, %s, %s', mode, _mode, _ptype, _vtype)

        for pos in range(1, len(_person_stages)+1):
            stage = _person_stages[pos]
            self.logger.debug('STAGE %d: %s', pos, pformat(stage))
            # findIntermodalRoute(self, fromEdge, toEdge, modes='', depart=-1.0,
            #                     routingMode=0, speed=-1.0, walkFactor=-1.0,
            #                     departPos=0.0, arrivalPos=-1073741824, departPosLat=0.0,
            #                     pType='', vType='', destStop='')
            if _current_depart_time is None:
                _current_depart_time = stage.start

            route = None

            if (stage.activity != 'Home' and
                    _vtype in self._conf['intermodalOptions']['vehicleAllowedParking']):
                # TRIP WITH PARKING REQUIREMENTS
                #  If the vtype is among the one that require parking, and we are not going home,
                #  look for a parking and build the additional walk back and forth.
                p_id, p_edge, _last_mile = self._env.find_closest_parking(stage.toEdge)
                if _last_mile:
                    route = traci.simulation.findIntermodalRoute(
                        stage.fromEdge, p_edge,
                        depart=_current_depart_time, walkFactor=.9,
                        modes=_mode, pType=_ptype, vType=_vtype)
                    if route and not isinstance(route, list):
                        # list in until SUMO 1.4.0 included, tuple onward
                        route = list(route)
                    if (sumoutils.is_valid_route(
                            mode, route,
                            self._conf['intermodalOptions']['vehicleAllowedParking']) and
                            route[-1].type == tc.STAGE_DRIVING):
                        route[-1].destStop = p_id
                        route[-1].arrivalPos = self._env.get_parking_position(p_id)
                        route.extend(_last_mile)
                    else:
                        route = None
                if route:
                    ## build the waiting to destination (if required)
                    if stage.duration:
                        route.append(self._generate_waiting_stage(stage))

                    ## build the walk back to the parking
                    walk_back = traci.simulation.findIntermodalRoute(
                        stage.toEdge, p_edge, walkFactor=.9, pType="pedestrian")
                    if route and not isinstance(route, list):
                        # list in until SUMO 1.4.0 included, tuple onward
                        route = list(route)
                    walk_back[-1].arrivalPos = self._env.get_parking_position(p_id)
                    route.extend(walk_back)

                    ## update the next stage to make it start from the parking
                    if pos + 1 in _person_stages:
                        _person_stages[pos+1] = _person_stages[pos+1]._replace(fromEdge=p_edge)

            elif _mode in ['public', 'taxi'] or _ptype in ['pedestrian']:
                # TRIP WITHOUT PARKING REQUIREMENTS, and no walk back and forth to mode change.
                #  If the _mode is among the one that are not generating the walk back and forth,
                #  the route provided by findIntermodalRoute.
                route = traci.simulation.findIntermodalRoute(
                    stage.fromEdge, stage.toEdge,
                    arrivalPos=stage.arrivalPos,
                    depart=_current_depart_time, walkFactor=.9,
                    modes=_mode, pType=_ptype, vType=_vtype)
                if not sumoutils.is_valid_route(
                    mode, route, self._conf['intermodalOptions']['vehicleAllowedParking']):
                    route = None
                if route and not isinstance(route, list):
                    # list in until SUMO 1.4.0 included, tuple onward
                    route = list(route)
                if route:
                    ## Set the arrival position in the edge
                    route[-1].arrivalPos = stage.arrivalPos
                    ## Add stop
                    if stage.duration:
                        route.append(self._generate_waiting_stage(stage))
            else:
                # TRIP WITHOUT PARKING REQUIREMENTS, BUT with walk back and forth to mode change.
                #  If the _mode  requires a mode change, the walk back and forth needs to be
                #  added after the waiting at destination due to the activity.
                route = traci.simulation.findIntermodalRoute(
                    stage.fromEdge, stage.toEdge,
                    arrivalPos=stage.arrivalPos,
                    depart=_current_depart_time, walkFactor=.9,
                    modes=_mode, pType=_ptype, vType=_vtype)
                if not sumoutils.is_valid_route(
                    mode, route, self._conf['intermodalOptions']['vehicleAllowedParking']):
                    route = None
                if route and not isinstance(route, list):
                    # list in until SUMO 1.4.0 included, tuple onward
                    route = list(route)
                if route:
                    ## Add stop
                    if stage.duration:
                        route.append(self._generate_waiting_stage(stage))
                    if not stage.final:
                        # If it's not the final stage,
                        #   check if it's required to add the walk back to the vehicle and
                        #   change the stage.fromEdge of the next stage.
                        self.logger.debug(
                            'Check walking back requirements for route \n%s', pformat(route))
                        if route[-2].type == tc.STAGE_WALKING:
                            # generate the walk back
                            from_edge = route[-2].edges[-1] # end ow the walking stage
                            to_edge = route[-3].edges[-1]    # end of the driving stage
                            arrival_position = self._env.get_arrival_pos_from_edge(
                                route[-3].edges[-1],
                                route[-3].arrivalPos) # parked vehicle
                            if arrival_position is None:
                                raise sagaexceptions.TripGenerationRouteError(
                                    'Arrival edge for the the walk back is too short.')
                            self.logger.debug(
                                'Walk back from %s to %s (%.2f --> %.2f)',
                                from_edge, to_edge, route[-3].arrivalPos, arrival_position)
                            walk_back = traci.simulation.findIntermodalRoute(
                                from_edge, to_edge, arrivalPos=arrival_position,
                                depart=(_current_depart_time + route[-1].travelTime),
                                walkFactor=.9, pType='pedestrian')
                            if not walk_back:
                                raise sagaexceptions.TripGenerationRouteError(
                                    'Route not found for the walk back from {} to {}.'.format(
                                        from_edge, to_edge))
                            self.logger.debug('Walk back: \n%s', pformat(walk_back))
                            if walk_back and not isinstance(walk_back, list):
                                # list in until SUMO 1.4.0 included, tuple onward
                                walk_back = list(walk_back)
                            route.extend(walk_back)
                            _person_stages[pos+1] = _person_stages[pos+1]._replace(
                                fromEdge=route[-1].edges[-1])

            ## Finalize the Journey
            if route and stage.final:
                ## fix the last stop with 1.0 duration
                if route[-1].type == tc.STAGE_WAITING:
                    route[-1].travelTime = 1.0
                    route[-1].cost = 1.0
                else:
                    raise sagaexceptions.TripGenerationInconsistencyError(
                        'Final route misses last waiting stage.', _person_stages)
                ## change the last triggered ride with LAST_STOP_PLACEHOLDER to fix the last stop
                if _mode not in ['public', 'taxi'] and _ptype not in ['pedestrian']:
                    _pos = len(route) - 1
                    while _pos >= 0:
                        if route[_pos].type == tc.STAGE_DRIVING:
                            route[_pos].destStop = self.LAST_STOP_PLACEHOLDER
                            break
                        _pos -= 1

            self.logger.debug('Route: \n%s', pformat(route))

            if route is None:
                raise sagaexceptions.TripGenerationRouteError(
                    'Route not found between {} and {}.'.format(stage.fromEdge, stage.toEdge))

            ## Add the stage to the full planned trip.
            for step in route:
                _current_depart_time += step.travelTime
                _person_steps.append(step)

            self.logger.debug('====================== Stage done. ======================')
        return _person_steps, _person_stages

    def _generate_waiting_stage(self, stage):
        """ Builds a STAGE_WAITING type of stage compatible with findIntermodalRoute. """
        wait = Stage(
            type=tc.STAGE_WAITING, description=stage.activity,
            edges='{}_0'.format(stage.toEdge), travelTime=stage.duration, cost=stage.duration,
            vType=None, line=None, destStop=None, length=None, intended=None, depart=None,
            departPos=None, arrivalPos=stage.arrivalPos)
        self.logger.debug('WAITING Stage: %s', pformat(wait))
        return wait

    ## ---------------------------------------------------------------------------------------- ##
    ##                                       TraCI to XML                                       ##
    ## ---------------------------------------------------------------------------------------- ##

    ROUTES_TPL = """<?xml version="1.0" encoding="UTF-8"?>

<!-- Generated with SUMO Activity-Based Mobility Generator [https://github.com/lcodeca/SUMOActivityGen] -->

<routes xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/routes_file.xsd"> {trips}
</routes>"""

    VEHICLE = """
    <vehicle id="{id}" type="{v_type}" depart="{depart}" departLane="best" arrivalPos="{arrival}">{route}{stop}
    </vehicle>"""

    ROUTE = """
        <route edges="{edges}"/>"""

    STOP_PARKING_TRIGGERED = """
        <stop parkingArea="{id}" triggered="true" expected="{person}"/>"""

    STOP_EDGE_TRIGGERED = """
        <stop lane="{lane}" parking="true" startPos="{start}" endPos="{end}" triggered="true" expected="{person}"/>""" #pylint: disable=c0301

    ONDEMAND_TRIGGERED = """
        <stop lane="{lane}" parking="true" startPos="{start}" endPos="{end}" duration="1.0"/>"""

    FINAL_STOP = """
        <stop lane="{lane}" duration="1.0"/>"""

    PERSON = """
    <person id="{id}" type="pedestrian" depart="{depart}">{stages}
    </person>"""

    WAIT = """
        <stop lane="{lane}" duration="{duration}" actType="{action}"/>"""

    WALK = """
        <walk edges="{edges}"/>"""

    WALK_W_ARRIVAL = """
        <walk edges="{edges}" arrivalPos="{arrival}"/>"""

    WALK_BUS = """
        <walk edges="{edges}" busStop="{busStop}"/>"""

    RIDE_BUS = """
        <ride busStop="{busStop}" lines="{lines}" intended="{intended}" depart="{depart}"/>"""

    RIDE_TRIGGERED = """
        <ride from="{from_edge}" to="{to_edge}" arrivalPos="{arrival}" lines="{vehicle_id}"/>"""

    VEHICLE_TRIGGERED = """
    <vehicle id="{id}" type="{v_type}" depart="triggered">{route}{stops}
    </vehicle>"""

    VEHICLE_TRIGGERED_DEPART = """
    <vehicle id="{id}" type="{v_type}" depart="triggered" departPos="{depart}">{route}{stops}
    </vehicle>"""

    def _generate_stop_position(self, edge, position):
        """ """
        start = position - self._conf['stopBufferDistance'] / 2.0
        end = position + self._conf['stopBufferDistance'] / 2.0
        if start <= 0.0:
            start = 1.0
            end = self._conf['stopBufferDistance']
        if position > end or end >= self._env.sumo_network.getEdge(edge).getLength():
            end = -1.0
        return start, end

    def _generate_sumo_trip_from_activitygen(self, person):
        """ Generate the XML string for SUMO route file from a person-trip. """
        self.logger.debug(' ............... _generate_sumo_trip_from_activitygen ............... ')
        self.logger.debug('\n%s', pformat(person))
        self.logger.debug(' ............... computation ............... ')
        complete_trip = ''
        triggered = ''
        _tr_id = '{}_tr'.format(person['id'])
        _triggered_vtype = ''
        _triggered_route = []
        _triggered_stops = ''
        stages = ''
        _last_arrival_pos = None
        _internal_consistency_check = []
        _waiting_stages = []
        for stage in person['stages']:
            self.logger.debug('Stage \n%s', pformat(stage))
            if stage.type == tc.STAGE_WAITING:
                _waiting_stages.append(stage)
                stages += self.WAIT.format(lane=stage.edges,
                                           duration=stage.travelTime,
                                           action=stage.description)
            elif stage.type == tc.STAGE_WALKING:
                if stage.destStop:
                    stages += self.WALK_BUS.format(
                        edges=' '.join(stage.edges), busStop=stage.destStop)
                else:
                    if stage.arrivalPos:
                        stages += self.WALK_W_ARRIVAL.format(
                            edges=' '.join(stage.edges), arrival=stage.arrivalPos)
                        _last_arrival_pos = stage.arrivalPos
                    else:
                        stages += self.WALK.format(edges=' '.join(stage.edges))
            elif stage.type == tc.STAGE_DRIVING:
                if stage.line != stage.intended:
                    # Public Transports
                    # !!! vType MISSING !!! line=164:0, intended=pt_bus_164:0.50
                    # intended is the transport id, so it must be different
                    stages += self.RIDE_BUS.format(
                        busStop=stage.destStop, lines=stage.line,
                        intended=stage.intended, depart=stage.depart)
                else:
                    # Triggered vehicle (vTyep == line == intended)
                    # vType=bicycle, line=bicycle, intended=bicycle
                    # vType=passenger, line=passenger, intended=passenger
                    # vType=motorcycle, line=motorcycle, intended=motorcycle
                    # vType=on-demand, line=on-demand, intended=on-demand
                    _ride_id = None
                    if stage.intended == 'on-demand':
                        _ride_id = 'taxi'
                    else:
                        ## add to the existing one
                        _ride_id = _tr_id
                        if _triggered_route:
                            ## check for contiguity
                            if _triggered_route[-1] != stage.edges[0]:
                                _error_message = 'Triggered vehicle {} has a broken route.'.format(
                                    stage.intended)
                                self.logger.warning(_error_message)
                                raise sagaexceptions.TripGenerationInconsistencyError(
                                    _error_message, pformat(person['stages']))
                            ## remove the duplicated edge
                            _triggered_route.extend(stage.edges[1:])
                        else:
                            ## nothing to be "fixed"
                            _triggered_route.extend(stage.edges)
                        self.logger.debug('_triggered_route: %s', _triggered_route)
                        _triggered_vtype = stage.vType
                        _stop = ''
                        self.logger.debug(
                            'travelTime: %f - destStop: %s', stage.travelTime, stage.destStop)
                        if stage.destStop == self.LAST_STOP_PLACEHOLDER:
                            self.logger.debug('Final stop reached.')
                            _stop = self.FINAL_STOP.format(
                                lane=self._env.get_stopping_lane(
                                    stage.edges[-1], self._vtype_to_vclasses[_triggered_vtype]))
                        else:
                            if stage.destStop and stage.vType in self._conf[
                                'intermodalOptions']['vehicleAllowedParking']:
                                self.logger.debug('Generate the stop in a parking area.')
                                _stop = self.STOP_PARKING_TRIGGERED.format(
                                    id=stage.destStop, person=person['id'])
                            else:
                                self.logger.debug('Generate the stop on the side of the road.')
                                start, end = self._generate_stop_position(
                                    stage.edges[-1], stage.arrivalPos)
                                self.logger.debug('Stop from %f to %f', start, end)
                                _stop = self.STOP_EDGE_TRIGGERED.format(
                                    lane=self._env.get_stopping_lane(
                                        stage.edges[-1], self._vtype_to_vclasses[_triggered_vtype]),
                                    person=person['id'], start=start, end=end)
                                self.logger.debug('_stop: %s', _stop)
                        _triggered_stops += _stop

                        self.logger.debug('_triggered_stops: %s', _triggered_stops)

                    stages += self.RIDE_TRIGGERED.format(
                        from_edge=stage.edges[0], to_edge=stage.edges[-1], vehicle_id=_ride_id,
                        arrival=stage.arrivalPos)

                    self.logger.debug('Stages: %s', stages)

            self.logger.debug(' -- Next stage -- ')

        ## fixing the personal triggered vehicles
        if _triggered_route:
            _route = self.ROUTE.format(edges=' '.join(_triggered_route))
            triggered += self.VEHICLE_TRIGGERED.format(
                id=_tr_id, v_type=_triggered_vtype, route=_route,
                stops=_triggered_stops, arrival='random', depart='random')

        ## waiting stages consistency test
        if not _waiting_stages:
            self.logger.warning('Person plan does not have any waiting stages.')
            raise sagaexceptions.TripGenerationInconsistencyError(
                'Person plan does not have any waiting stages.',
                person['stages'])

        ## result
        complete_trip += triggered
        complete_trip += self.PERSON.format(
            id=person['id'], depart=person['depart'], stages=stages)

        self.logger.debug('Complete trip: \n%s', complete_trip)
        return complete_trip

    ## ---------------------------------------------------------------------------------------- ##
    ##                                Saving trips to files                                     ##
    ## ---------------------------------------------------------------------------------------- ##

    def _saving_trips_to_files(self):
        """ Saving all the trips to files divided by slice. """
        for name, dict_trips in self._all_trips.items():
            filename = '{}{}.rou.xml'.format(self._conf['outputPrefix'], name)
            with open(filename, 'w') as tripfile:
                all_trips = ''
                for time in sorted(dict_trips.keys()):
                    for person in dict_trips[time]:
                        all_trips += person['string']

                tripfile.write(self.ROUTES_TPL.format(trips=all_trips))
            self.logger.info('Saved %s', filename)

    def _saving_trips_to_single_file(self):
        """ Saving all the trips into a single file. """
        ## Sort (by time) all the slice into one
        merged_trips = collections.defaultdict(list)
        for dict_trips in self._all_trips.values():
            for time in sorted(dict_trips.keys()):
                for person in dict_trips[time]:
                    merged_trips[time].append(person['string'])

        filename = '{}.merged.rou.xml'.format(self._conf['outputPrefix'])
        with open(filename, 'w') as tripfile:
            all_trips = ''
            for time in sorted(merged_trips.keys()):
                for person in merged_trips[time]:
                    all_trips += person

            tripfile.write(self.ROUTES_TPL.format(trips=all_trips))
            self.logger.info('Saved %s', filename)

def main(cmd_args):
    """ Person Trip Activity-based Mobility Generation with PoIs and TAZ. """
    args = get_options(cmd_args)

    ## ========================              PROFILER              ======================== ##
    if args.profiling:
        profiler = cProfile.Profile()
        profiler.enable()
    ## ========================              PROFILER              ======================== ##

    print('Loading configuration file {}.'.format(args.config))
    MobilityGenerator(
        json.loads(open(args.config).read()),
        profiling=args.profiling).generate()

    ## ========================              PROFILER              ======================== ##
    if args.profiling:
        profiler.disable()
        results = io.StringIO()
        pstats.Stats(profiler, stream=results).sort_stats('cumulative').print_stats(25)
        print(results.getvalue())
    ## ========================              PROFILER              ======================== ##

    print('Done.')

if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3

""" SUMO Activity-Based Mobility Generator - Exceptions

    Author: Lara CODECA

    This program and the accompanying materials are made available under the
    terms of the Eclipse Public License 2.0 which is available at
    http://www.eclipse.org/legal/epl-2.0.
"""

import logging
import json
from pprint import pformat
import sys

LOGGER = logging.getLogger('ActivityGen')

class TripGenerationGenericError(Exception):
    """
    During the trip generation, various erroroneous states can be reached.
    """
    def __init__(self, message=None):
        """ Init the error message. """
        super().__init__()
        self.message = message
        if self.message:
            LOGGER.debug(self.message)

class TripGenerationActivityError(TripGenerationGenericError):
    """
    During the generation from the activity chains, various erroroneous states can be reached.
    """
    def __init__(self, message, activity=None):
        """ Init the error message. """
        super().__init__()
        self.message = message
        self.activity = activity
        LOGGER.debug(self.message)
        if self.activity is not None:
            with open('TripGenerationActivityError.log', 'a') as openfile:
                # openfile.write(message + '\n')
                json.dump(
                    {
                        'msg': message,
                        'activity': activity,
                    },
                    openfile, indent=2)
                # openfile.write(pformat(activity) + '\n')

class TripGenerationRouteError(TripGenerationGenericError):
    """
    During the step by step generation of the trip, it is possible to reach a state in which
        some of the chosen locations are impossible to reach.
    """
    def __init__(self, message, route=None):
        """ Init the error message. """
        super().__init__()
        self.message = message
        self.route = route
        LOGGER.debug(self.message)
        if self.route is not None:
            with open('TripGenerationRouteError.log', 'a') as openfile:
                # openfile.write(message + '\n')
                json.dump(
                    {
                        'msg': message,
                        'route': route,
                    },
                    openfile, indent=2)
                # openfile.write(pformat(route) + '\n')

class TripGenerationInconsistencyError(TripGenerationGenericError):
    """
    During the step by step generation of the trip, it is possible to reach a state in which
        some of the chosen modes are impossible to be used in that order.
    """
    def __init__(self, message, plan=None):
        """ Init the error message. """
        super().__init__()
        self.message = message
        self.plan = plan
        LOGGER.debug(self.message)
        if self.plan is not None:
            with open('TripGenerationInconsistencyError.log', 'a') as openfile:
                # openfile.write(message + '\n')
                json.dump(
                    {
                        'msg': message,
                        'plan': plan,
                    },
                    openfile, indent=2)
                # openfile.write(pformat(plan) + '\n')



# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2012-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    miscutils.py
# @author  Jakob Erdmann
# @author  Michael Behrisch
# @date    2012-05-08

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import sys
import time
import os
import math
import colorsys
import socket
import random
# needed for backward compatibility
from .statistics import Statistics, geh, uMax, uMin, round  # noqa


def benchmark(func):
    """
    decorator for timing a function
    """
    def benchmark_wrapper(*args, **kwargs):
        started = time.time()
        now = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.localtime())
        print('function %s called at %s' % (func.__name__, now))
        sys.stdout.flush()
        result = func(*args, **kwargs)
        print('function %s finished after %f seconds' %
              (func.__name__, time.time() - started))
        sys.stdout.flush()
        return result
    return benchmark_wrapper


class working_dir:
    """
    temporarily change working directory using 'with' statement
    """

    def __init__(self, dir):
        self.dir = dir
        self.origdir = os.getcwd()

    def __enter__(self):
        os.chdir(self.dir)

    def __exit__(self, type, value, traceback):
        os.chdir(self.origdir)


class Colorgen:
    DISTINCT = [
        (0.17, 1.0, 0.5),
        (0.0, 0.9, 1.0),
        (0.35, 0.67, 0.71),
        (0.14, 0.9, 1.0),
        (0.56, 1.0, 0.78),
        (0.07, 0.8, 0.96),
        (0.79, 0.83, 0.71),
        (0.5, 0.71, 0.94),
        (0.84, 0.79, 0.94),
        (0.2, 0.76, 0.96),
        (0.0, 0.24, 0.98),
        (0.5, 1.0, 0.5),
        (0.77, 0.25, 1.0),
        (0.09, 0.76, 0.67),
        (0.15, 0.22, 1.0),
        (0.0, 1.0, 0.5),
        (0.38, 0.33, 1.0),
        (0.67, 1.0, 0.5),
    ]

    def __init__(self, hsv, cycleLength=10.67):
        self.hsv = hsv
        self.cycle = [int(random.random() * 256) for x in self.hsv]
        self.cycleOffset = int(round(256 / cycleLength))
        self.distinctIndex = 0

    def get_value(self, opt, index):
        if opt == 'random':
            return random.random()
        elif opt == 'cycle':
            # the 255 below is intentional to get all color values when cycling long enough
            self.cycle[index] = (self.cycle[index] + self.cycleOffset) % 255
            return self.cycle[index] / 255.0
        elif opt == 'cycle':
            # the 255 below is intentional to get all color values when cycling long enough
            self.cycle[index] = (self.cycle[index] + self.cycleOffset) % 255
            return self.cycle[index] / 255.0
        elif opt == 'distinct':
            if index == 0:
                self.distinctIndex = (self.distinctIndex + 1) % len(self.DISTINCT)
            return self.DISTINCT[self.distinctIndex][index]
        else:
            return float(opt)

    def floatTuple(self):
        """return color as a tuple of floats each in [0,1]"""
        return colorsys.hsv_to_rgb(*[self.get_value(o, i) for i, o in enumerate(self.hsv)])

    def byteTuple(self):
        """return color as a tuple of bytes each in [0,255]"""
        return tuple([int(round(255 * x)) for x in self.floatTuple()])

    def __call__(self):
        """return constant or randomized rgb-color string"""
        return ','.join(map(str, self.byteTuple()))


def getFreeSocketPort(numTries=10):
    for _ in range(numTries):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(('', 0))
            p = s.getsockname()[1]
            s.close()
            return p
        except socket.error:
            pass
    return None


def getSocketStream(port, mode='rb'):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("localhost", port))
    s.listen(1)
    conn, _ = s.accept()
    return conn.makefile(mode)


# euclidean distance between two coordinates in the plane
def euclidean(a, b):
    return math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)


def humanReadableTime(seconds):
    result = ""
    ds = 3600 * 24
    if seconds > ds:
        result = "%s:" % int(seconds / ds)
        seconds = seconds % ds
    result += "%02i:" % int(seconds / 3600)
    seconds = seconds % 3600
    result += "%02i:" % int(seconds / 60)
    seconds = seconds % 60
    if seconds == int(seconds):
        seconds = int(seconds)
    result += "%02i" % seconds
    return result


def parseTime(t, factor=1):
    try:
        return float(t) * factor
    except ValueError:
        pass
    # prepended zero is ignored if the date value already contains days
    days, hours, minutes, seconds = ([0] + list(map(float, t.split(':'))))[-4:]
    return 3600 * 24 * days + 3600 * hours + 60 * minutes + seconds


def parseBool(val):
    # see data/xsd/baseTypes:boolType
    return val in ["true", "True", "x", "1", "yes", "on"]



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2008-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    dailyNetedit.py
# @author  Michael Behrisch
# @author  Jakob Erdmann
# @author  Laura Bieker
# @date    2008

"""
Does the nightly git pull on the windows server and the visual
studio build. The script is also used for the meso build.
Some paths especially for the names of the texttest output dirs are
hard coded into this script.
"""
from __future__ import absolute_import
from __future__ import print_function
import datetime
import optparse
import os
import glob
import subprocess
import sys

import status

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import sumolib  # noqa

BINARIES = ("netedit",)


def killall(debugSuffix):
    bins = set([name + debugSuffix + ".exe" for name in BINARIES])
    for taskline in subprocess.check_output(["tasklist", "/nh"]).splitlines():
        task = taskline.split()
        if task and task[0] in bins:
            subprocess.call(["taskkill", "/f", "/im", task[0]])
            bins.remove(task[0])


def runTests(options, env, gitrev, log, debugSuffix=""):
    prefix = env["FILEPREFIX"] + debugSuffix
    env["SUMO_BATCH_RESULT"] = os.path.join(options.rootDir, prefix + "batch_result")
    env["SUMO_REPORT"] = os.path.join(options.remoteDir, prefix + "report")
    env["TEXTTEST_TMP"] = os.path.join(options.rootDir, prefix + "texttesttmp")
    env["TEXTTEST_HOME"] = os.path.join(options.rootDir, options.testsDir)
    if not os.path.exists(env["SUMO_REPORT"]):
        os.makedirs(env["SUMO_REPORT"])
    killall(debugSuffix)
    for name in BINARIES:
        binary = os.path.join(options.rootDir, options.binDir, name + debugSuffix + ".exe")
        if os.path.exists(binary):
            env[name.upper() + "_BINARY"] = binary
    ttBin = "texttestc.py"
    today = datetime.date.today()
    tasks = sorted(glob.glob(os.path.join(env["TEXTTEST_HOME"], "netedit", "testsuite.netedit.daily.*")))
    taskID = os.path.basename(tasks[today.toordinal() % len(tasks)])[10:]
    cmd = [ttBin, "-b", prefix, "-a", taskID, "-name", "%sr%s" % (today.strftime("%d%b%y"), gitrev)]
    subprocess.call(cmd, env=env, stdout=log, stderr=subprocess.STDOUT, shell=True)
    subprocess.call([ttBin, "-b", env["FILEPREFIX"], "-coll"], env=env,
                    stdout=log, stderr=subprocess.STDOUT, shell=True)
    killall(debugSuffix)


optParser = optparse.OptionParser()
optParser.add_option("-r", "--root-dir", dest="rootDir",
                     default=r"D:\Sumo", help="root for git and log output")
optParser.add_option("-s", "--suffix", default="", help="suffix to the fileprefix")
optParser.add_option("-b", "--bin-dir", dest="binDir", default=r"git\bin",
                     help="directory containg the binaries, relative to the root dir")
optParser.add_option("-t", "--tests-dir", dest="testsDir", default=r"git\tests",
                     help="directory containg the tests, relative to the root dir")
optParser.add_option("-m", "--remote-dir", dest="remoteDir", default="S:\\daily",
                     help="directory to move the results to")
optParser.add_option("-p", "--python", help="path to python interpreter to use")
(options, args) = optParser.parse_args()

env = os.environ
if "SUMO_HOME" not in env:
    env["SUMO_HOME"] = os.path.dirname(
        os.path.dirname(os.path.dirname(__file__)))
env["PYTHON"] = "python"
env["SMTP_SERVER"] = "smtprelay.dlr.de"
msvcVersion = "msvc12"

platform = "x64"
env["FILEPREFIX"] = msvcVersion + options.suffix + platform
prefix = os.path.join(options.remoteDir, env["FILEPREFIX"])
testLog = prefix + "NeteditTest.log"
gitrev = sumolib.version.gitDescribe()
with open(testLog, 'a') as log:
    status.printLog("Running tests.", log)
    runTests(options, env, gitrev, log)



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2020-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    osmTaxiStop.py
# @author  Michael Behrisch
# @date    2020-10-30

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import io
import argparse
sys.path.append(os.path.join(os.environ["SUMO_HOME"], "tools"))
import sumolib  # noqa
import osmBuild  # noqa


def main(options):
    if options.net_file is None:
        osmBuild.build(["-f", options.osm_file])
        net = sumolib.net.readNet("osm.net.xml")
    else:
        net = sumolib.net.readNet(options.net_file)
    count = 0
    with io.open(options.output_file, "w") as output:
        sumolib.xml.writeHeader(output, root="additional")
        for n in sumolib.xml.parse(options.osm_file, "node"):
            name = None
            bestLane = None
            point = None
            length = options.length
            if n.tag:
                for t in n.tag:
                    if t.k == "capacity":
                        try:
                            length = max(int(t.v) * 7.5, length)
                        except ValueError:
                            pass
                    if t.k == "name":
                        name = t.v
                    if t.k == "amenity" and t.v == "taxi":
                        point = net.convertLonLat2XY(float(n.lon), float(n.lat))
                        for lane, _ in sorted(net.getNeighboringLanes(*point, r=options.radius), key=lambda i: i[1]):
                            if lane.getLength() > options.length and lane.allows(options.vclass):
                                bestLane = lane
                                break
                if bestLane:
                    pos = sumolib.geomhelper.polygonOffsetWithMinimumDistanceToPoint(point, bestLane.getShape())
                    endPos = min(lane.getLength(), max(length, pos + length / 2))
                    nameAttr = 'name="%s" ' % name if name else ""
                    print('    <%s id="%s_%s" %slane="%s" startPos="%.2f" endPos="%.2f"/>' %
                          (options.type, options.type, count, nameAttr, bestLane.getID(), endPos - length, endPos),
                          file=output)
                    count += 1
        print(u"</additional>", file=output)


argParser = sumolib.options.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
argParser.add_argument("-s", "--osm-file", help="read OSM file from FILE (mandatory)", metavar="FILE")
argParser.add_argument("-n", "--net-file", help="read SUMO net from FILE", metavar="FILE")
argParser.add_argument("-o", "--output-file", help="write stopping places to the output FILE", metavar="FILE",
                       default="stopping_places.add.xml")
argParser.add_argument("-t", "--type", help="stopping place type", default="chargingStation")
argParser.add_argument("-r", "--radius", type=float, help="radius for edge finding", default=20.)
argParser.add_argument("-l", "--length", type=float, help="(minimum) length of the stopping place", default=20.)
argParser.add_argument("--vclass", help="which vehicle class should be allowed", default="passenger")
options = argParser.parse_args()
if not options.osm_file:
    argParser.print_help()
    sys.exit()
main(options)
